#include <thread>
#include <iostream>
#include <mutex>

class atomic_counter {
public:
    int value() {
        std::unique_lock<std::mutex> l(m_mutex);
        return m_counter;
    }
    
    template <typename T>
    auto with(T function) {
        std::unique_lock<std::mutex> l(m_mutex);
        return function(m_counter);
    }
    
private:
    int m_counter = 0;
    std::mutex m_mutex;
};

int main() {
    for (int i = 0; i < 1000000; ++i) {
        atomic_counter counter;

        auto thread_func = [&counter]() {
            while (counter.with([](int& c) {
                if (c < 5) {
                    c++;
                    return true;
                }
                return false;
            })) {
                // do something
            }
        };

        std::thread t1(thread_func);
        std::thread t2(thread_func);
        t1.join();
        t2.join();

        if (counter.value() != 5) {
            std::cout << counter.value() << '\n';
        }
    }
}
